from typing import Union, Optional, List, Iterator, Iterable, IO
from copy import deepcopy

from gpx_lite.gpxtrackpoint import GPXTrackPoint


class GPXTrackSegment:

    __slots__ = '_points'

    def __init__(self, points: Optional[List[GPXTrackPoint]]=None):
        self._points = points if points else []

    def __repr__(self)-> str:
        return 'GPXTrackSegment(%s)(points=%s)' % \
               (len(self._points), self._points)

    def __len__(self)-> int:
        return len(self._points)

    def __getitem__(self, key:Union[int, slice])-> \
            Union[GPXTrackPoint, List[GPXTrackPoint]]:
        if isinstance(key, int):
            return self.points[key]
        elif isinstance(key, slice):
            return self.points[key.start:key.stop:key.step]
        else:
            raise TypeError('Index must be int, not {}'.
                            format(type(key).__name__))

    def __iter__(self)->Iterator[GPXTrackPoint]:
        return iter(self.points)

    def __contains__(self, item: GPXTrackPoint)->bool:
        return item in self.points

    @property
    def points(self)->List[GPXTrackPoint]:
        return self._points

    @points.setter
    def points(self, points: List[GPXTrackPoint]):
        self._points = points

    def append(self, item: GPXTrackPoint):
        self._points.append(item)

    def extend(self, items: Iterable[GPXTrackPoint]):
        self._points.extend(items)

    def remove(self, item: GPXTrackPoint):
        self._points.remove(item)

    def get_points_no(self)->int:
        """
        Gets the number of points in segment.
        """
        return len(self._points)

    def to_xml(self, fh:IO)->None:
        fh.write('\n<trkseg>')
        for pt in self._points:
            pt.to_xml(fh)
        fh.write('\n</trkseg>')

    def sort_by_time(self)->None:
        self._points.sort(key=lambda pt: pt._time)

    def clone(self):
        return deepcopy(self)


if __name__ == '__main__':

    x = 50.0164596
    y = 14.4547907
    p1 = GPXTrackPoint(x, y, '2017-11-22T07:03:36Z')
    p3 = GPXTrackPoint(y,y, '2617-11-13T08:11:09Z')
    p2 = GPXTrackPoint(x, x, '2017-12-02T07:03:36Z')
    p4 = GPXTrackPoint(y,x, '1717-11-13T08:11:09Z')
    seg = GPXTrackSegment()
    print('Empty segment: ', seg)
    seg.append(p1)
    print('Length,  1 point: ', len(seg))
    seg.extend([p2,p3])
    print('0th element: ', seg[0])
    print('Segment with 3 points: ', seg)
    print('seg.points: ', seg.points)
    print('Point in segment: %s, not in segment: %s ' % (
        p1 in seg, p4 in seg))
    print('Slice: ', seg[1:2:2])
    print('Iterator:')
    for p in seg:
        print(p)




from typing import IO, Callable, List
from xml.etree.ElementTree import ElementTree, iterparse

from gpx_lite.gpx import GPX
from gpx_lite.gpxtrack import GPXTrack
from gpx_lite.gpxtrackpoint import GPXTrackPoint
from gpx_lite.gpxtracksegment import GPXTrackSegment
from gpx_lite.utils import parse_xml


class GPXParser:
    """
    Parser for gpx tracks.

    Args:
       file:  file handler

    Usage:

        import gpx-lite as parser
                 ...
        with open(filename, 'r') as gpx_file:
            gpx:GPX = parser.parse(gpx_file)
                  or
            gpx:GPX = parser.iterparse(gpx_file)

    """

    __slots__ = ('_gpx', '_source')

    def __init__(self, file: IO)->None:
        self._source: IO = file
        self._gpx: GPX = GPX()

    def parse(self, xml_parser: Callable = parse_xml)->GPX:
        """

        :param xml_parser:
        :return:
        """
        xml: ElementTree = xml_parser(self._source.read())
        for trk in xml.iterfind('trk'):
            new_track = GPXTrack()
            try:
                new_track.name = trk.find('name').text
            except AttributeError:
                pass
            try:
                new_track.number = trk.find('number').text
            except AttributeError:
                pass
            for seg in trk.iterfind('trkseg'):
                new_segment = GPXTrackSegment()
                for point in seg.iterfind('trkpt'):
                    values = point.attrib
                    try:
                        point.attrib['time'] = point.find('time').text
                    except AttributeError:
                        point.attrib['time'] = None
                    new_point = GPXTrackPoint(float(values['lat']),
                                              float(values['lon']),
                                              values['time'])
                    new_segment.append(new_point)
                new_track.append(new_segment)
            self._gpx.append(new_track)
        return self._gpx

    def iterparse(self)->GPX:
        points: List[GPXTrackPoint] = []
        segments: List[GPXTrackSegment]= []
        tracks: List[GPXTrack] = []
        name: str = ''
        number: str = ''
        time: str = ''

        for _, elem in iterparse(self._source):
            if 'name' in elem.tag:
                name = elem.text
            elif 'number' in elem.tag:
                number = elem.text
            elif 'time' in elem.tag:
                time = elem.text
            elif 'trkpt' in elem.tag:
                points.append(GPXTrackPoint(float(elem.attrib['lat']),
                                            float(elem.attrib['lon']),
                                            time))
            elif 'trkseg' in elem.tag:
                segments.append(GPXTrackSegment([p for p in points]))
                points.clear()
            elif 'trk' in elem.tag:
                tracks.append(GPXTrack(name, number, [s for s in segments]))
                segments.clear()
            elem.clear()
        self._gpx = GPX(tracks=tracks)
        return self._gpx


if __name__ == '__main__':
    fn = '/home/olga/Documents/GPX/test1.gpx'
    with open(fn, 'r') as xml_file:
        parser = GPXParser(xml_file)
        gpx = parser.parse()
    print(gpx)
    for track in gpx:
        print(track)
        for seg in track:
            print(seg)
            for pt in seg:
                print(pt)
                print(pt.latitude, pt.longitude,pt.time)

    fn2 = '/home/olga/Documents/GPX/traces-raw.gpx'
    with open(fn2, 'r') as fh:
        parser = GPXParser(fh)
        gpx = parser.iterparse()
    with open('/home/olga/Documents/GPX/save_raw.gpx', 'w') as fh2:
        print(gpx.to_xml(fh2))
    print(gpx)



from typing import Optional, List, Union, Iterator, Iterable, IO
from copy import deepcopy

from gpx_lite.gpxtrack import GPXTrack


class GPX:
    """
    Container for GPXTracks. This class represents the root element of gpx file.
    All constructor arguments are optional.

    Attributes:
        version: version of gpx schema
        creator: application that created the gpx
        tracks: list of tracks in this gpx
    """

    __slots__ = ('_version', '_creator', '_tracks')

    def __init__(self, version: Optional[str]=None,
                 creator: Optional[str]=None,
                 tracks: Optional[List[GPXTrack]]=None):
        """
        :param version: version of gpx schema
        :param creator: application that created the data
        :param tracks: list of tracks
        """
        self._version: Optional[str] = version
        self._creator: Optional[str] = creator
        self._tracks: List[GPXTrack] = tracks if tracks else []

    def __repr__(self)->str:
        return '<GPX [..%s tracks..]>' % len(self._tracks)

    def __getitem__(self, key: Union[int, slice])-> \
            Union[GPXTrack, List[GPXTrack]]:

        if isinstance(key, int):
            return self._tracks[key]
        elif isinstance(key, slice):
            return self._tracks[key.start:key.stop:key.step]
        else:
            raise TypeError('Index must be int, not {}'.
                            format(type(key).__name__))

    def __len__(self)->int:
        return len(self._tracks)

    def __contains__(self, item: GPXTrack)->bool:
        return item in self._tracks

    def __iter__(self)->Iterator[GPXTrack]:
        return iter(self._tracks)

    @property
    def tracks(self)->List[GPXTrack]:
        return self._tracks

    @tracks.setter
    def tracks(self, items: List[GPXTrack]):
        self._tracks = items

    @property
    def version(self)->str:
        return self._version

    @version.setter
    def version(self, version: str):
        self._version = version

    @property
    def creator(self)->str:
        return self._creator

    @creator.setter
    def creator(self, creator: str):
        self._creator = creator

    def append(self, item: GPXTrack):
        self._tracks.append(item)

    def extend(self, items:Iterable[GPXTrack]):
        self._tracks.extend(items)

    def remove(self, item: GPXTrack):
        self._tracks.remove(item)

    def to_xml(self, fh: IO)->None:
        version: str = self.version if self.version else '1.1'
        creator: str = self.creator if self.creator else 'gpx-lite.py'
        version_ns: str = version.replace('.', '/')
        result: List[str] = ['<?xml version="1.0" encoding="UTF-8"?>',
                             '\n<gpx xmlns="http://www.topografix.com/GPX/%s" ' % version_ns,
                             'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" ',
                             'xsi:schemaLocation="http://www.topografix.com/GPX/%s ' % version_ns,
                             'http://www.topografix.com/GPX/%s/gpx.xsd" ' % version_ns,
                             'version="%s" ' % version,
                             'creator="%s">' % creator]
        for string in result:
            fh.write(string)

        for track in self._tracks:
            track.to_xml(fh)
        fh.write('\n</gpx>')

    def clone(self)->'GPX':
        return deepcopy(self)


if __name__ == '__main__':

    from gpx_lite.gpxtrackpoint import GPXTrackPoint as TrackPoint
    from gpx_lite.gpxtracksegment import GPXTrackSegment as TrackSegment

    x = 50.0164596
    y = 14.4547907
    p1 = TrackPoint(x, y, '2017-11-22T11:03:36Z')
    p2 = TrackPoint(y, x,'2017-11-22T08:03:36Z')
    p3 = TrackPoint(y,y, '2017-11-13T05:11:09Z')
    p4 = TrackPoint(x, x, '2017-11-22T09:03:36Z')
    seg1 = TrackSegment([p1, p2, p3])
    seg2 = TrackSegment([p2, p3, p4])
    seg3 = TrackSegment([p4, p1])
    track1 = GPXTrack('800003627_337', '0', [seg1, seg2, seg3])
    track2 = GPXTrack('800003627_908', None, [seg2, seg3])
    track3 = GPXTrack(None, '2', [TrackSegment([p4])])
    print('Track with name, number, 3 segments: ',track1)
    print('Track with name,no number, 2 segments: ',track2)
    print('Track with no name, number, 1 segment: ', track3)

    gpx = GPX('1.0','gpx.py -- https://github.com/tkrajina/gpxpy')
    print('Empty gpx with version and creator: ', gpx)
    gpx.append(track1)
    print('1 track added: ', gpx)
    gpx.extend([track2,track3])
    print('2 more tracks added, len = ', gpx)

    print('Slice: ', gpx[0:1])
    print('third track: ', gpx[2])

    print('Iterator')
    for t in gpx:
        print(t)

    gpx.remove(track3)
    print('.tracks after 1 track removed: ', gpx.tracks)





from typing import Dict, Tuple, Iterator, List, Union, Iterable
from tqdm import tqdm
from os import path, listdir

from gpx_lite.gpx import GPX
from gpx_lite.gpxtrackpoint import GPXTrackPoint
from gpx_lite.gpxtracksegment import GPXTrackSegment
from gpx_lite.gpxtrack import GPXTrack

import roadmaptools.inout
from tqdm import tqdm


def read_data_from_csv(filepath: str, gpx: GPX, id_dict: Dict[str, int])->None:
    """
    Loads data from .csv file to gpx object.

    :param filepath:
    :param gpx:
    :param id_dict: dictionary, maps car_id to track position in gpx tracks list.
    :return:
    """
    iter: Iterable = roadmaptools.inout.load_csv(filepath)
    for row in iter:
        car_id = row[1]
        point = GPXTrackPoint(float(row[3]), float(row[4]), row[5])
        if car_id not in id_dict.keys():
            num = len(id_dict)
            id_dict[car_id] = num
            gpx.append(GPXTrack(car_id, str(num), [GPXTrackSegment([point,]),]))
        else:
            num = id_dict[car_id]
            gpx[num][0].append(point)



if __name__=='__main__':

    dir:str = '/home/olga/Documents/GPX/csv_data_liftago'
    filenames: List[str] = [path.join(dir, fn) for fn in listdir(dir)]
    gpx:GPX = GPX()
    id_dict:Dict[str, int] = {}
    for fn in tqdm(filenames, total=len(filenames), desc="Loading csv..."):
        read_data_from_csv(fn, gpx, id_dict)

    for tr in gpx:
        tr[0].sort_by_time()

    with open('/home/olga/Documents/GPX/liftago.gpx', 'w') as fh:
        gpx.to_xml(fh)




from datetime import datetime
from typing import Callable, List, IO

from gpx_lite.utils import parse_time


class GPXTrackPoint:
    """
    Track point, leaf element of gpx

    Attributes:
        latitude: float
        longitude: float
        time:   datetime

    """
    __slots__ = ('_lat', '_lon', '_time')

    def __init__(self, lat: float, lon: float, time: str)->None:
        """
        :param lat: point latitude
        :param lon: point longitude
        :param time: time
        """
        self._lat: float = lat
        self._lon: float = lon
        self._time: str = time

    def __repr__(self)->str:
        return '<GPXTrackPoint(%f, %f, %s)>' % \
               (self._lat, self._lon, self._time)

    def __str__(self)->str:
        return 'trkpt:%s %s %s' % (self._lat, self._lon, self._time)

    @property
    def latitude(self)->float:
        return self._lat

    @property
    def longitude(self) -> float:
        return self._lon

    @property
    def time(self, parser: Callable = parse_time)->datetime:
        return parser(self._time)

    def to_xml(self, fh:IO)->None:
        result: List[str] = ['\n<trkpt lat="%f" lon="%f">' % (self._lat, self._lon),
                             '\n<time>', self._time, '</time>', '\n</trkpt>']
        for string in result:
            fh.write(string)


if __name__ == '__main__':
    p0 = GPXTrackPoint(70.016978, 41.3749454,
                       '2016-12-22T11:50:02Z')
    print('p0: ',p0)
    p1 = GPXTrackPoint(70.024596, 41.4547907,
                       '2017-02-22T07:25:02Z')
    print('p1:', p1)
    print('p1.latitude=%s, p1.longitude=%s, p1.time=%s' %
          (p1.latitude, p1.longitude, p1.time))


